package eu.jonahbauer.raytracing.math;

import eu.jonahbauer.raytracing.scene.Hittable;
import org.jetbrains.annotations.NotNull;

import java.util.Comparator;
import java.util.List;
import java.util.Optional;

/**
 * An axis-aligned bounding box.
 */
public record AABB(@NotNull Vec3 min, @NotNull Vec3 max) {
    public static final AABB EMPTY = new AABB(Vec3.ZERO, Vec3.ZERO);
    public static final Comparator<AABB> X_AXIS = Comparator.comparing(AABB::min, Comparator.comparingDouble(Vec3::x));
    public static final Comparator<AABB> Y_AXIS = Comparator.comparing(AABB::min, Comparator.comparingDouble(Vec3::y));
    public static final Comparator<AABB> Z_AXIS = Comparator.comparing(AABB::min, Comparator.comparingDouble(Vec3::z));

    public AABB {
        var a = min;
        var b = max;
        min = Vec3.min(a, b);
        max = Vec3.max(a, b);
    }

    public AABB(@NotNull Range x, @NotNull Range y, @NotNull Range z) {
        this(new Vec3(x.min(), y.min(), z.min()), new Vec3(x.max(), y.max(), z.max()));
    }

    public static @NotNull Optional<AABB> getBoundingBox(@NotNull List<? extends @NotNull Hittable> objects) {
        var bbox = (AABB) null;
        for (var object : objects) {
            bbox = bbox == null ? object.getBoundingBox() : bbox.expand(object.getBoundingBox());
        }
        return Optional.ofNullable(bbox);
    }

    /**
     * {@return the range of x values}
     */
    public @NotNull Range x() {
        return new Range(min.x(), max.x());
    }

    /**
     * {@return the range of y values}
     */
    public @NotNull Range y() {
        return new Range(min.y(), max.y());
    }

    /**
     * {@return the range of z values}
     */
    public @NotNull Range z() {
        return new Range(min.z(), max.z());
    }

    /**
     * {@return the center of this bounding box}
     */
    public @NotNull Vec3 center() {
        return Vec3.average(min, max, 2);
    }

    /**
     * Expands this bounding box to include the other bounding box.
     * @param box a bounding box
     * @return the expanded bounding box
     */
    public @NotNull AABB expand(@NotNull AABB box) {
        return new AABB(Vec3.min(this.min, box.min), Vec3.max(this.max, box.max));
    }

    /**
     * Tests whether the {@code ray} intersects this bounding box withing the {@code range}
     * @param ray a ray
     * @param range a range of valid {@code t}s
     * @return {@code true} iff the ray intersects this bounding box, {@code false} otherwise
     */
    public boolean hit(@NotNull Ray ray, @NotNull Range range) {
        var invDirection = ray.getInvDirection();
        var negInvOrigin = ray.getNegInvOrigin();

        var tminX = Math.fma(min.x(), invDirection.x(), negInvOrigin.x());
        var tminY = Math.fma(min.y(), invDirection.y(), negInvOrigin.y());
        var tminZ = Math.fma(min.z(), invDirection.z(), negInvOrigin.z());

        var tmaxX = Math.fma(max.x(), invDirection.x(), negInvOrigin.x());
        var tmaxY = Math.fma(max.y(), invDirection.y(), negInvOrigin.y());
        var tmaxZ = Math.fma(max.z(), invDirection.z(), negInvOrigin.z());

        var tlmax = max(
                Math.min(tminX, tmaxX),
                Math.min(tminY, tmaxY),
                Math.min(tminZ, tmaxZ)
        );
        var tumin = min(
                Math.max(tminX, tmaxX),
                Math.max(tminY, tmaxY),
                Math.max(tminZ, tmaxZ)
        );

        return tlmax < tumin && tumin >= range.min() && tlmax <= range.max();
    }

    private static double max(double a, double b, double c) {
        return Math.max(a, Math.max(b, c));
    }

    private static double min(double a, double b, double c) {
        return Math.min(a, Math.min(b, c));
    }

    /**
     * Computes the {@code t} values of the intersections of a ray with the axis-aligned planes through a point.
     * @param corner the point
     * @param origin the origin point of the ray
     * @param invDirection the {@linkplain Vec3#inv() inverted} direction of the ray
     * @return a three-element array of the {@code t} values of the intersection with the yz-, xz- and xy-plane through {@code corner}
     */
    public static double @NotNull[] intersect(@NotNull Vec3 corner, @NotNull Vec3 origin, @NotNull Vec3 invDirection) {
        return new double[] {
                (corner.x() - origin.x()) * invDirection.x(),
                (corner.y() - origin.y()) * invDirection.y(),
                (corner.z() - origin.z()) * invDirection.z(),
        };
    }
}
