package eu.jonahbauer.raytracing.render.material;

import eu.jonahbauer.raytracing.math.Ray;
import eu.jonahbauer.raytracing.math.Vec3;
import eu.jonahbauer.raytracing.render.spectrum.Spectra;
import eu.jonahbauer.raytracing.render.texture.Texture;
import eu.jonahbauer.raytracing.scene.HitResult;
import org.jetbrains.annotations.NotNull;

import java.util.Objects;
import java.util.Optional;
import java.util.random.RandomGenerator;

public record DielectricMaterial(@NotNull RefractiveIndex ri, @NotNull Texture texture) implements Material {
    public DielectricMaterial {
        Objects.requireNonNull(ri, "ri");
        Objects.requireNonNull(texture, "texture");
    }

    public DielectricMaterial(@NotNull RefractiveIndex ri) {
        this(ri, Spectra.WHITE);
    }

    public DielectricMaterial(double ri) {
        this(new ConstantRefractiveIndex(ri), Spectra.WHITE);
    }

    public DielectricMaterial(double ri, @NotNull Texture texture) {
        this(new ConstantRefractiveIndex(ri), texture);
    }

    @Override
    public @NotNull Optional<ScatterResult> scatter(@NotNull Ray ray, @NotNull HitResult hit, @NotNull RandomGenerator random) {
        var ri = switch (this.ri) {
            case ConstantRefractiveIndex(var x) -> x;
            case RefractiveIndex x -> x.get(ray.lambda().collapse());
        };
        if (hit.isFrontFace()) ri = 1 / ri;

        var cosTheta = Math.min(- ray.direction().unit().dot(hit.normal()), 1.0);
        var reflectance = reflectance(cosTheta, ri);
        var reflect = reflectance > random.nextDouble();

        var newDirection = (reflect ? Optional.<Vec3>empty() : Vec3.refract(ray.direction(), hit.normal(), ri))
                .orElseGet(() -> Vec3.reflect(ray.direction(), hit.normal()));

        var attenuation = texture.get(hit);
        return Optional.of(new SpecularScatterResult(attenuation, ray.with(hit, newDirection)));
    }

    private double reflectance(double cos, double ri) {
        // use schlick's approximation for reflectance
        var r0 = (1 - ri) / (1 + ri);
        r0 = r0 * r0;
        return r0 + (1 - r0) * (1 - cos) * (1 - cos) * (1 - cos) * (1 - cos) * (1 - cos);
    }

    @FunctionalInterface
    public interface RefractiveIndex {
        double get(double lambda);
    }

    public record ConstantRefractiveIndex(double ri) implements RefractiveIndex {

        @Override
        public double get(double lambda) {
            return ri;
        }
    }

    public record SellmeierRefractiveIndex(
            double B1, double B2, double B3,
            double C1, double C2, double C3
    ) implements RefractiveIndex {

        @Override
        public double get(double lambda) {
            var l2 = lambda * lambda * 1E-6; // square and convert to µm
            var x = 1 + B1 * l2 / (l2 - C1)
                      + B2 * l2 / (l2 - C2)
                      + B3 * l2 / (l2 - C3);
            return Math.sqrt(x);
        }
    }
}
